@@ -4,8 +4,10 @@ from io import StringIO
 from typing import Optional
 
 import numpy as np
+
 from gym import Env, spaces
 from gym.envs.toy_text.utils import categorical_sample
+from gym.utils.renderer import Renderer
 
 UP = 0
 RIGHT = 1
@@ -41,8 +43,10 @@ class CliffWalkingEnv(Env):
     - 3: move left
 
     ### Observations
-    There are 3x12 + 1 possible states. In fact, the agent cannot be at the cliff, nor at the goal (as this results the end of episode). They remain all the positions of the first 3 rows plus the bottom-left cell.
-    The observation is simply the current position encoded as [flattened index](https://numpy.org/doc/stable/reference/generated/numpy.unravel_index.html).
+    There are 3x12 + 1 possible states. In fact, the agent cannot be at the cliff, nor at the goal
+    (as this results the end of episode). They remain all the positions of the first 3 rows plus the bottom-left cell.
+    The observation is simply the current position encoded as
+    [flattened index](https://numpy.org/doc/stable/reference/generated/numpy.unravel_index.html).
 
     ### Reward
     Each time step incurs -1 reward, and stepping into the cliff incurs -100 reward.
@@ -59,7 +63,7 @@ class CliffWalkingEnv(Env):
 
     metadata = {"render_modes": ["human", "ansi"], "render_fps": 4}
 
-    def __init__(self):
+    def __init__(self, render_mode: Optional[str] = None):
         self.shape = (4, 12)
         self.start_state_index = np.ravel_multi_index((3, 0), self.shape)
 
@@ -88,12 +92,11 @@ class CliffWalkingEnv(Env):
         self.observation_space = spaces.Discrete(self.nS)
         self.action_space = spaces.Discrete(self.nA)
 
-    def _limit_coordinates(self, coord):
-        """
-        Prevent the agent from falling out of the grid world
-        :param coord:
-        :return:
-        """
+        self.render_mode = render_mode
+        self.renderer = Renderer(self.render_mode, self._render)
+
+    def _limit_coordinates(self, coord: np.ndarray) -> np.ndarray:
+        """Prevent the agent from falling out of the grid world."""
         coord[0] = min(coord[0], self.shape[0] - 1)
         coord[0] = max(coord[0], 0)
         coord[1] = min(coord[1], self.shape[1] - 1)
@@ -101,11 +104,14 @@ class CliffWalkingEnv(Env):
         return coord
 
     def _calculate_transition_prob(self, current, delta):
-        """
-        Determine the outcome for an action. Transition Prob is always 1.0.
-        :param current: Current position on the grid as (row, col)
-        :param delta: Change in position for transition
-        :return: (1.0, new_state, reward, done)
+        """Determine the outcome for an action. Transition Prob is always 1.0.
+
+        Args:
+            current: Current position on the grid as (row, col)
+            delta: Change in position for transition
+
+        Returns:
+            Tuple of ``(1.0, new_state, reward, terminated)``
         """
         new_position = np.array(current) + np.array(delta)
         new_position = self._limit_coordinates(new_position).astype(int)
@@ -114,16 +120,17 @@ class CliffWalkingEnv(Env):
             return [(1.0, self.start_state_index, -100, False)]
 
         terminal_state = (self.shape[0] - 1, self.shape[1] - 1)
-        is_done = tuple(new_position) == terminal_state
-        return [(1.0, new_state, -1, is_done)]
+        is_terminated = tuple(new_position) == terminal_state
+        return [(1.0, new_state, -1, is_terminated)]
 
     def step(self, a):
         transitions = self.P[self.s][a]
         i = categorical_sample([t[0] for t in transitions], self.np_random)
-        p, s, r, d = transitions[i]
+        p, s, r, t = transitions[i]
         self.s = s
         self.lastaction = a
-        return (int(s), r, d, {"prob": p})
+        self.renderer.render_step()
+        return (int(s), r, t, False, {"prob": p})
 
     def reset(
         self,
@@ -135,12 +142,21 @@ class CliffWalkingEnv(Env):
         super().reset(seed=seed)
         self.s = categorical_sample(self.initial_state_distrib, self.np_random)
         self.lastaction = None
+        self.renderer.reset()
+        self.renderer.render_step()
         if not return_info:
             return int(self.s)
         else:
             return int(self.s), {"prob": 1}
 
     def render(self, mode="human"):
+        if self.render_mode is not None:
+            return self.renderer.get_renders()
+        else:
+            return self._render(mode)
+
+    def _render(self, mode):
+        assert mode in self.metadata["render_modes"]
         outfile = StringIO() if mode == "ansi" else sys.stdout
 
         for s in range(self.nS):
