@@ -1,31 +1,34 @@
-from typing import Optional, Union, List
+"""An async vector environment."""
+import multiprocessing as mp
+import sys
+import time
+from copy import deepcopy
+from enum import Enum
+from typing import List, Optional, Sequence, Tuple, Union
 
 import numpy as np
-import multiprocessing as mp
-import time
-import sys
-from enum import Enum
-from copy import deepcopy
 
+import gym
 from gym import logger
-from gym.logger import warn
-from gym.vector.vector_env import VectorEnv
+from gym.core import ObsType
 from gym.error import (
     AlreadyPendingCallError,
-    NoAsyncCallError,
     ClosedEnvironmentError,
     CustomSpaceError,
+    NoAsyncCallError,
 )
+from gym.utils.step_api_compatibility import step_api_compatibility
 from gym.vector.utils import (
-    create_shared_memory,
-    create_empty_array,
-    write_to_shared_memory,
-    read_from_shared_memory,
-    concatenate,
-    iterate,
     CloudpickleWrapper,
     clear_mpi_env_vars,
+    concatenate,
+    create_empty_array,
+    create_shared_memory,
+    iterate,
+    read_from_shared_memory,
+    write_to_shared_memory,
 )
+from gym.vector.vector_env import VectorEnv
 
 __all__ = ["AsyncVectorEnv"]
 
@@ -38,69 +41,13 @@ class AsyncState(Enum):
 
 
 class AsyncVectorEnv(VectorEnv):
-    """Vectorized environment that runs multiple environments in parallel. It
-    uses `multiprocessing`_ processes, and pipes for communication.
+    """Vectorized environment that runs multiple environments in parallel.
 
-    Parameters
-    ----------
-    env_fns : iterable of callable
-        Functions that create the environments.
+    It uses ``multiprocessing`` processes, and pipes for communication.
 
-    observation_space : :class:`gym.spaces.Space`, optional
-        Observation space of a single environment. If ``None``, then the
-        observation space of the first environment is taken.
-
-    action_space : :class:`gym.spaces.Space`, optional
-        Action space of a single environment. If ``None``, then the action space
-        of the first environment is taken.
-
-    shared_memory : bool
-        If ``True``, then the observations from the worker processes are
-        communicated back through shared variables. This can improve the
-        efficiency if the observations are large (e.g. images).
-
-    copy : bool
-        If ``True``, then the :meth:`~AsyncVectorEnv.reset` and
-        :meth:`~AsyncVectorEnv.step` methods return a copy of the observations.
-
-    context : str, optional
-        Context for `multiprocessing`_. If ``None``, then the default context is used.
-
-    daemon : bool
-        If ``True``, then subprocesses have ``daemon`` flag turned on; that is, they
-        will quit if the head process quits. However, ``daemon=True`` prevents
-        subprocesses to spawn children, so for some environments you may want
-        to have it set to ``False``.
-
-    worker : callable, optional
-        If set, then use that worker in a subprocess instead of a default one.
-        Can be useful to override some inner vector env logic, for instance,
-        how resets on done are handled.
-
-    Warning
-    -------
-    :attr:`worker` is an advanced mode option. It provides a high degree of
-    flexibility and a high chance to shoot yourself in the foot; thus,
-    if you are writing your own worker, it is recommended to start from the code
-    for ``_worker`` (or ``_worker_shared_memory``) method, and add changes.
-
-    Raises
-    ------
-    RuntimeError
-        If the observation space of some sub-environment does not match
-        :obj:`observation_space` (or, by default, the observation space of
-        the first sub-environment).
-
-    ValueError
-        If :obj:`observation_space` is a custom space (i.e. not a default
-        space in Gym, such as :class:`~gym.spaces.Box`, :class:`~gym.spaces.Discrete`,
-        or :class:`~gym.spaces.Dict`) and :obj:`shared_memory` is ``True``.
-
-    Example
-    -------
-
-    .. code-block::
+    Example::
 
+        >>> import gym
         >>> env = gym.vector.AsyncVectorEnv([
         ...     lambda: gym.make("Pendulum-v0", g=9.81),
         ...     lambda: gym.make("Pendulum-v0", g=1.62)
@@ -112,15 +59,46 @@ class AsyncVectorEnv(VectorEnv):
 
     def __init__(
         self,
-        env_fns,
-        observation_space=None,
-        action_space=None,
-        shared_memory=True,
-        copy=True,
-        context=None,
-        daemon=True,
-        worker=None,
+        env_fns: Sequence[callable],
+        observation_space: Optional[gym.Space] = None,
+        action_space: Optional[gym.Space] = None,
+        shared_memory: bool = True,
+        copy: bool = True,
+        context: Optional[str] = None,
+        daemon: bool = True,
+        worker: Optional[callable] = None,
+        new_step_api: bool = False,
     ):
+        """Vectorized environment that runs multiple environments in parallel.
+
+        Args:
+            env_fns: Functions that create the environments.
+            observation_space: Observation space of a single environment. If ``None``,
+                then the observation space of the first environment is taken.
+            action_space: Action space of a single environment. If ``None``,
+                then the action space of the first environment is taken.
+            shared_memory: If ``True``, then the observations from the worker processes are communicated back through
+                shared variables. This can improve the efficiency if the observations are large (e.g. images).
+            copy: If ``True``, then the :meth:`~AsyncVectorEnv.reset` and :meth:`~AsyncVectorEnv.step` methods
+                return a copy of the observations.
+            context: Context for `multiprocessing`_. If ``None``, then the default context is used.
+            daemon: If ``True``, then subprocesses have ``daemon`` flag turned on; that is, they will quit if
+                the head process quits. However, ``daemon=True`` prevents subprocesses to spawn children,
+                so for some environments you may want to have it set to ``False``.
+            worker: If set, then use that worker in a subprocess instead of a default one.
+                Can be useful to override some inner vector env logic, for instance, how resets on termination or truncation are handled.
+            new_step_api: If True, step method returns 2 bools - terminated, truncated, instead of 1 bool - done
+
+        Warnings: worker is an advanced mode option. It provides a high degree of flexibility and a high chance
+            to shoot yourself in the foot; thus, if you are writing your own worker, it is recommended to start
+            from the code for ``_worker`` (or ``_worker_shared_memory``) method, and add changes.
+
+        Raises:
+            RuntimeError: If the observation space of some sub-environment does not match observation_space
+                (or, by default, the observation space of the first sub-environment).
+            ValueError: If observation_space is a custom space (i.e. not a default space in Gym,
+                such as gym.spaces.Box, gym.spaces.Discrete, or gym.spaces.Dict) and shared_memory is True.
+        """
         ctx = mp.get_context(context)
         self.env_fns = env_fns
         self.shared_memory = shared_memory
@@ -137,6 +115,7 @@ class AsyncVectorEnv(VectorEnv):
             num_envs=len(env_fns),
             observation_space=observation_space,
             action_space=action_space,
+            new_step_api=new_step_api,
         )
 
         if self.shared_memory:
@@ -193,6 +172,14 @@ class AsyncVectorEnv(VectorEnv):
         self._check_spaces()
 
     def seed(self, seed=None):
+        """Seeds the vector environments.
+
+        Args:
+            seed: The seeds use with the environments
+
+        Raises:
+            AlreadyPendingCallError: Calling `seed` while waiting for a pending call to complete
+        """
         super().seed(seed=seed)
         self._assert_is_running()
         if seed is None:
@@ -218,18 +205,20 @@ class AsyncVectorEnv(VectorEnv):
         return_info: bool = False,
         options: Optional[dict] = None,
     ):
-        """Send the calls to :obj:`reset` to each sub-environment.
+        """Send calls to the :obj:`reset` methods of the sub-environments.
 
-        Raises
-        ------
-        ClosedEnvironmentError
-            If the environment was closed (if :meth:`close` was previously called).
+        To get the results of these calls, you may invoke :meth:`reset_wait`.
 
-        AlreadyPendingCallError
-            If the environment is already waiting for a pending call to another
-            method (e.g. :meth:`step_async`). This can be caused by two consecutive
-            calls to :meth:`reset_async`, with no call to :meth:`reset_wait` in
-            between.
+        Args:
+            seed: List of seeds for each environment
+            return_info: If to return information
+            options: The reset option
+
+        Raises:
+            ClosedEnvironmentError: If the environment was closed (if :meth:`close` was previously called).
+            AlreadyPendingCallError: If the environment is already waiting for a pending call to another
+                method (e.g. :meth:`step_async`). This can be caused by two consecutive
+                calls to :meth:`reset_async`, with no call to :meth:`reset_wait` in between.
         """
         self._assert_is_running()
 
@@ -259,37 +248,26 @@ class AsyncVectorEnv(VectorEnv):
 
     def reset_wait(
         self,
-        timeout=None,
+        timeout: Optional[Union[int, float]] = None,
         seed: Optional[int] = None,
         return_info: bool = False,
         options: Optional[dict] = None,
-    ):
-        """
-        Parameters
-        ----------
-        timeout : int or float, optional
-            Number of seconds before the call to `reset_wait` times out. If
-            `None`, the call to `reset_wait` never times out.
-        seed: ignored
-        options: ignored
+    ) -> Union[ObsType, Tuple[ObsType, List[dict]]]:
+        """Waits for the calls triggered by :meth:`reset_async` to finish and returns the results.
 
-        Returns
-        -------
-        element of :attr:`~VectorEnv.observation_space`
-            A batch of observations from the vectorized environment.
-        infos : list of dicts containing metadata
+        Args:
+            timeout: Number of seconds before the call to `reset_wait` times out. If `None`, the call to `reset_wait` never times out.
+            seed: ignored
+            return_info: If to return information
+            options: ignored
 
-        Raises
-        ------
-        ClosedEnvironmentError
-            If the environment was closed (if :meth:`close` was previously called).
+        Returns:
+            A tuple of batched observations and list of dictionaries
 
-        NoAsyncCallError
-            If :meth:`reset_wait` was called without any prior call to
-            :meth:`reset_async`.
-
-        TimeoutError
-            If :meth:`reset_wait` timed out.
+        Raises:
+            ClosedEnvironmentError: If the environment was closed (if :meth:`close` was previously called).
+            NoAsyncCallError: If :meth:`reset_wait` was called without any prior call to :meth:`reset_async`.
+            TimeoutError: If :meth:`reset_wait` timed out.
         """
         self._assert_is_running()
         if self._state != AsyncState.WAITING_RESET:
@@ -309,8 +287,10 @@ class AsyncVectorEnv(VectorEnv):
         self._state = AsyncState.DEFAULT
 
         if return_info:
-            results, infos = zip(*results)
-            infos = list(infos)
+            infos = {}
+            results, info_data = zip(*results)
+            for i, info in enumerate(info_data):
+                infos = self._add_info(infos, info, i)
 
             if not self.shared_memory:
                 self.observations = concatenate(
@@ -328,24 +308,18 @@ class AsyncVectorEnv(VectorEnv):
 
             return deepcopy(self.observations) if self.copy else self.observations
 
-    def step_async(self, actions):
+    def step_async(self, actions: np.ndarray):
         """Send the calls to :obj:`step` to each sub-environment.
 
-        Parameters
-        ----------
-        actions : element of :attr:`~VectorEnv.action_space`
-            Batch of actions.
+        Args:
+            actions: Batch of actions. element of :attr:`~VectorEnv.action_space`
 
-        Raises
-        ------
-        ClosedEnvironmentError
-            If the environment was closed (if :meth:`close` was previously called).
-
-        AlreadyPendingCallError
-            If the environment is already waiting for a pending call to another
-            method (e.g. :meth:`reset_async`). This can be caused by two consecutive
-            calls to :meth:`step_async`, with no call to :meth:`step_wait` in
-            between.
+        Raises:
+            ClosedEnvironmentError: If the environment was closed (if :meth:`close` was previously called).
+            AlreadyPendingCallError: If the environment is already waiting for a pending call to another
+                method (e.g. :meth:`reset_async`). This can be caused by two consecutive
+                calls to :meth:`step_async`, with no call to :meth:`step_wait` in
+                between.
         """
         self._assert_is_running()
         if self._state != AsyncState.DEFAULT:
@@ -359,40 +333,21 @@ class AsyncVectorEnv(VectorEnv):
             pipe.send(("step", action))
         self._state = AsyncState.WAITING_STEP
 
-    def step_wait(self, timeout=None):
+    def step_wait(
+        self, timeout: Optional[Union[int, float]] = None
+    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, List[dict]]:
         """Wait for the calls to :obj:`step` in each sub-environment to finish.
 
-        Parameters
-        ----------
-        timeout : int or float, optional
-            Number of seconds before the call to :meth:`step_wait` times out. If
-            ``None``, the call to :meth:`step_wait` never times out.
+        Args:
+            timeout: Number of seconds before the call to :meth:`step_wait` times out. If ``None``, the call to :meth:`step_wait` never times out.
 
-        Returns
-        -------
-        observations : element of :attr:`~VectorEnv.observation_space`
-            A batch of observations from the vectorized environment.
+        Returns:
+             The batched environment step information, (obs, reward, terminated, truncated, info) or (obs, reward, done, info) depending on new_step_api
 
-        rewards : :obj:`np.ndarray`, dtype :obj:`np.float_`
-            A vector of rewards from the vectorized environment.
-
-        dones : :obj:`np.ndarray`, dtype :obj:`np.bool_`
-            A vector whose entries indicate whether the episode has ended.
-
-        infos : list of dict
-            A list of auxiliary diagnostic information dicts from sub-environments.
-
-        Raises
-        ------
-        ClosedEnvironmentError
-            If the environment was closed (if :meth:`close` was previously called).
-
-        NoAsyncCallError
-            If :meth:`step_wait` was called without any prior call to
-            :meth:`step_async`.
-
-        TimeoutError
-            If :meth:`step_wait` timed out.
+        Raises:
+            ClosedEnvironmentError: If the environment was closed (if :meth:`close` was previously called).
+            NoAsyncCallError: If :meth:`step_wait` was called without any prior call to :meth:`step_async`.
+            TimeoutError: If :meth:`step_wait` timed out.
         """
         self._assert_is_running()
         if self._state != AsyncState.WAITING_STEP:
@@ -407,10 +362,21 @@ class AsyncVectorEnv(VectorEnv):
                 f"The call to `step_wait` has timed out after {timeout} second(s)."
             )
 
-        results, successes = zip(*[pipe.recv() for pipe in self.parent_pipes])
+        observations_list, rewards, terminateds, truncateds, infos = [], [], [], [], {}
+        successes = []
+        for i, pipe in enumerate(self.parent_pipes):
+            result, success = pipe.recv()
+            obs, rew, terminated, truncated, info = step_api_compatibility(result, True)
+
+            successes.append(success)
+            observations_list.append(obs)
+            rewards.append(rew)
+            terminateds.append(terminated)
+            truncateds.append(truncated)
+            infos = self._add_info(infos, info, i)
+
         self._raise_if_errors(successes)
         self._state = AsyncState.DEFAULT
-        observations_list, rewards, dones, infos = zip(*results)
 
         if not self.shared_memory:
             self.observations = concatenate(
@@ -419,25 +385,29 @@ class AsyncVectorEnv(VectorEnv):
                 self.observations,
             )
 
-        return (
-            deepcopy(self.observations) if self.copy else self.observations,
-            np.array(rewards),
-            np.array(dones, dtype=np.bool_),
-            infos,
+        return step_api_compatibility(
+            (
+                deepcopy(self.observations) if self.copy else self.observations,
+                np.array(rewards),
+                np.array(terminateds, dtype=np.bool_),
+                np.array(truncateds, dtype=np.bool_),
+                infos,
+            ),
+            self.new_step_api,
+            True,
         )
 
-    def call_async(self, name, *args, **kwargs):
-        """
-        Parameters
-        ----------
-        name : string
-            Name of the method or property to call.
+    def call_async(self, name: str, *args, **kwargs):
+        """Calls the method with name asynchronously and apply args and kwargs to the method.
 
-        *args
-            Arguments to apply to the method call.
+        Args:
+            name: Name of the method or property to call.
+            *args: Arguments to apply to the method call.
+            **kwargs: Keyword arguments to apply to the method call.
 
-        **kwargs
-            Keywoard arguments to apply to the method call.
+        Raises:
+            ClosedEnvironmentError: If the environment was closed (if :meth:`close` was previously called).
+            AlreadyPendingCallError: Calling `call_async` while waiting for a pending call to complete
         """
         self._assert_is_running()
         if self._state != AsyncState.DEFAULT:
@@ -451,19 +421,19 @@ class AsyncVectorEnv(VectorEnv):
             pipe.send(("_call", (name, args, kwargs)))
         self._state = AsyncState.WAITING_CALL
 
-    def call_wait(self, timeout=None):
-        """
-        Parameters
-        ----------
-        timeout : int or float, optional
-            Number of seconds before the call to `step_wait` times out. If
-            `None` (default), the call to `step_wait` never times out.
+    def call_wait(self, timeout: Optional[Union[int, float]] = None) -> list:
+        """Calls all parent pipes and waits for the results.
 
-        Returns
-        -------
-        results : list
-            List of the results of the individual calls to the method or
-            property for each environment.
+        Args:
+            timeout: Number of seconds before the call to `step_wait` times out.
+                If `None` (default), the call to `step_wait` never times out.
+
+        Returns:
+            List of the results of the individual calls to the method or property for each environment.
+
+        Raises:
+            NoAsyncCallError: Calling `call_wait` without any prior call to `call_async`.
+            TimeoutError: The call to `call_wait` has timed out after timeout second(s).
         """
         self._assert_is_running()
         if self._state != AsyncState.WAITING_CALL:
@@ -484,17 +454,18 @@ class AsyncVectorEnv(VectorEnv):
 
         return results
 
-    def set_attr(self, name, values):
-        """
-        Parameters
-        ----------
-        name : string
-            Name of the property to be set in each individual environment.
+    def set_attr(self, name: str, values: Union[list, tuple, object]):
+        """Sets an attribute of the sub-environments.
 
-        values : list, tuple, or object
-            Values of the property to be set to. If `values` is a list or
-            tuple, then it corresponds to the values for each individual
-            environment, otherwise a single value is set for all environments.
+        Args:
+            name: Name of the property to be set in each individual environment.
+            values: Values of the property to be set to. If ``values`` is a list or
+                tuple, then it corresponds to the values for each individual
+                environment, otherwise a single value is set for all environments.
+
+        Raises:
+            ValueError: Values must be a list or tuple with length equal to the number of environments.
+            AlreadyPendingCallError: Calling `set_attr` while waiting for a pending call to complete.
         """
         self._assert_is_running()
         if not isinstance(values, (list, tuple)):
@@ -518,25 +489,19 @@ class AsyncVectorEnv(VectorEnv):
         _, successes = zip(*[pipe.recv() for pipe in self.parent_pipes])
         self._raise_if_errors(successes)
 
-    def close_extras(self, timeout=None, terminate=False):
-        """Close the environments & clean up the extra resources
-        (processes and pipes).
+    def close_extras(
+        self, timeout: Optional[Union[int, float]] = None, terminate: bool = False
+    ):
+        """Close the environments & clean up the extra resources (processes and pipes).
 
-        Parameters
-        ----------
-        timeout : int or float, optional
-            Number of seconds before the call to :meth:`close` times out. If ``None``,
-            the call to :meth:`close` never times out. If the call to :meth:`close`
-            times out, then all processes are terminated.
+        Args:
+            timeout: Number of seconds before the call to :meth:`close` times out. If ``None``,
+                the call to :meth:`close` never times out. If the call to :meth:`close`
+                times out, then all processes are terminated.
+            terminate: If ``True``, then the :meth:`close` operation is forced and all processes are terminated.
 
-        terminate : bool
-            If ``True``, then the :meth:`close` operation is forced and all processes
-            are terminated.
-
-        Raises
-        ------
-        TimeoutError
-            If :meth:`close` timed out.
+        Raises:
+            TimeoutError: If :meth:`close` timed out.
         """
         timeout = 0 if terminate else timeout
         try:
@@ -614,7 +579,7 @@ class AsyncVectorEnv(VectorEnv):
 
         num_errors = self.num_envs - sum(successes)
         assert num_errors > 0
-        for _ in range(num_errors):
+        for i in range(num_errors):
             index, exctype, value = self.error_queue.get()
             logger.error(
                 f"Received the following error from Worker-{index}: {exctype.__name__}: {value}"
@@ -623,10 +588,12 @@ class AsyncVectorEnv(VectorEnv):
             self.parent_pipes[index].close()
             self.parent_pipes[index] = None
 
-        logger.error("Raising the last exception back to the main process.")
-        raise exctype(value)
+            if i == num_errors - 1:
+                logger.error("Raising the last exception back to the main process.")
+                raise exctype(value)
 
     def __del__(self):
+        """On deleting the object, checks that the vector environment is closed."""
         if not getattr(self, "closed", True) and hasattr(self, "_state"):
             self.close(terminate=True)
 
@@ -639,7 +606,7 @@ def _worker(index, env_fn, pipe, parent_pipe, shared_memory, error_queue):
         while True:
             command, data = pipe.recv()
             if command == "reset":
-                if "return_info" in data and data["return_info"] == True:
+                if "return_info" in data and data["return_info"] is True:
                     observation, info = env.reset(**data)
                     pipe.send(((observation, info), True))
                 else:
@@ -647,11 +614,17 @@ def _worker(index, env_fn, pipe, parent_pipe, shared_memory, error_queue):
                     pipe.send((observation, True))
 
             elif command == "step":
-                observation, reward, done, info = env.step(data)
-                if done:
-                    info["terminal_observation"] = observation
+                (
+                    observation,
+                    reward,
+                    terminated,
+                    truncated,
+                    info,
+                ) = step_api_compatibility(env.step(data), True)
+                if terminated or truncated:
+                    info["final_observation"] = observation
                     observation = env.reset()
-                pipe.send(((observation, reward, done, info), True))
+                pipe.send(((observation, reward, terminated, truncated, info), True))
             elif command == "seed":
                 env.seed(data)
                 pipe.send((None, True))
@@ -703,7 +676,7 @@ def _worker_shared_memory(index, env_fn, pipe, parent_pipe, shared_memory, error
         while True:
             command, data = pipe.recv()
             if command == "reset":
-                if "return_info" in data and data["return_info"] == True:
+                if "return_info" in data and data["return_info"] is True:
                     observation, info = env.reset(**data)
                     write_to_shared_memory(
                         observation_space, index, observation, shared_memory
@@ -716,14 +689,20 @@ def _worker_shared_memory(index, env_fn, pipe, parent_pipe, shared_memory, error
                     )
                     pipe.send((None, True))
             elif command == "step":
-                observation, reward, done, info = env.step(data)
-                if done:
-                    info["terminal_observation"] = observation
+                (
+                    observation,
+                    reward,
+                    terminated,
+                    truncated,
+                    info,
+                ) = step_api_compatibility(env.step(data), True)
+                if terminated or truncated:
+                    info["final_observation"] = observation
                     observation = env.reset()
                 write_to_shared_memory(
                     observation_space, index, observation, shared_memory
                 )
-                pipe.send(((None, reward, done, info), True))
+                pipe.send(((None, reward, terminated, truncated, info), True))
             elif command == "seed":
                 env.seed(data)
                 pipe.send((None, True))
